  ' FTP server
  ' TassyJim
  ' Feb 2021
  ' Server requires ESP32 running latest AT firmware (V2.1.0.0)
  ' For handshaking, GPIO14 (RTS) needs to be connected to any user defined IO pin (default pin is 36)
  ' If you are using an ESP32-S2 series chip, this will be different.
  ' I included a 1k resistor in series to prevent issues if other software uses the RTS pin as an output.
  ' Don't use pins 37 or 40
  '
  ' Passive mode not supported
  ' tested with Total Commander on Windows and Android.
  ' also tested with gFTP on Linux Mint.
  ' a file "FTPserver.ini" is written to the current folder on program exit.
  ' this can be edited to change default values
  ' you can also change the default values in prefData:
  '
  ' a baudrate of 460800 gives file saving @ 38kB/s and 19.9kB/s sending
  ' a higher baudrate doesn't result in any improvement due to limitations in the ESP32 firmware.
  ' Currently, the ESP8266 cannot be used with standard firmware
  '
  ' if the program fails to exit cleanly, the ini file might not get updated. 
  ' If you have changed the baudrate, you may have to manually edit the ini file to regain control
  '
  OPTION EXPLICIT
  OPTION DEFAULT INTEGER
  ' Global variables
  CONST qt$=CHR$(34)
  CONST crlf$=CHR$(13)+CHR$(10)
  '
  ' all FTP messages only really need the 3 digit code. the rest is for human benefit.
  CONST welcomeTxt$ = "220----Welcome to my CMM2"+crlf$+"220      Please be nice"+crlf$
  CONST features$="211-Extensions supported:"+crlf$+"SIZE"+crlf$+"211 End."+crlf$ '"PASV"+crlf$+
  CONST systype$ = "215 UNIX emulated by MMBasic"+crlf$
  CONST unknownCMD$ = "504 Unknown command"+crlf$
  CONST nowASCII$ = "200 TYPE is now ASCII"+crlf$
  CONST nowBINARY$ = "200 TYPE is now 8-bit binary"+crlf$
  CONST acceptTF$ = "150 Accepted data connection"+crlf$
  CONST successTF$ = "226 Transfer complete"+crlf$
  CONST passOK$ = "230 OK. Current restricted directory is /"+crlf$
  CONST finishList$ = "226 "
  CONST goodbye$ = "221 Goodbye. "+crlf$+"221 Logout."+crlf$
  CONST activePortOK$ ="200 PORT command successful."+crlf$
  CONST startTF$ = "150 Opening BINARY mode data connection for file transfer."+crlf$
  CONST mkdirOK$ = "257 Directory created successfully."+crlf$
  CONST namefrom$ = "350 RNFR accepted - file exists, ready for destination"+crlf$
  CONST nameto$ = "250 File successfully renamed."+crlf$
  '
  DIM OldFolder$
  DIM prefs$(15,2) LENGTH 32 ' array for preferences.
  DIM comport$, ComSpeed$, NTPdata$
  DIM SSID$=qt$+"mySSID"+qt$
  DIM SSIDPassword$=qt$+"myPASSWORD"+qt$
  DIM INTEGER MTU = 2048
  DIM esp32 = 1 ' 1 = ESP32, 0 = ESP8266
  DIM doreset = 1, RTSpin = 36
  DIM inTxt$ ' global string for input buffer
  DIM INTEGER quit, showDebug, pkt, verbose = 1
  DIM ink$, a$ 'variables used by main loop only
  
  DIM INTEGER x, filecount, quitting, CIPRECVMODE, startTime
  DIM stationIP$, CMDx$, user$, clientIP$, ntpTime$, CMDtxt$, FTPtimeout$,para2$
  DIM INTEGER clientPort
  
  openprefs "FTPserver.ini"
  configure
  DIM INTEGER obuff(MTU\8 + 1) 'used to buffer writes to the internet to optimise speed
  SETPIN RTSpin, DIN
  PRINT comPort$+ComSpeed$+",8192"
  OPEN comPort$+ComSpeed$+",8192" AS #1
  PAUSE 2000 'wait for ESP32 to wake up afer power on
  CLOSE #1 'clear the buffer
  OPEN comPort$+ComSpeed$+",8192" AS #1
  OldFolder$ = CWD$ ' so we can restore starting folder on exit
  IF RIGHT$(OldFolder$,1)<>"/" AND RIGHT$(OldFolder$,1)<>"\" THEN OldFolder$ = OldFolder$+"/"
  ESPcommand("AT") ' test ESP for responce
  IF OK()THEN
    PRINT "ESP alive!"
  ELSE
    PAUSE 200
    ESPcommand("AT")
    IF OK()THEN
      PRINT "ESP alive!"
    ELSE
      PRINT inTxt$
    ENDIF
  ENDIF
  
  IF doreset THEN
    x = ESPreset()
  ELSE
    x = 1
  ENDIF
  
  IF x = 1 THEN ESPstart
  
  PAUSE 600 ' give NTP time to do it's thing
  
  settimefromNTP
  
  showStatus
  '> main loop
  DO
    ink$ = INKEY$
    SELECT CASE ink$
      CASE CHR$(27)
        quit = 1
      CASE "B","b" ' change Baudrate immediately
        changeBaud
      CASE "R","r" ' do ESP reset
        x = ESPreset()
        IF x = 1 THEN ESPstart
      CASE "H" ,"h" ' help and status
        showStatus
      CASE "D","d"
        showDebug = 1 - showDebug
        IF showDebug THEN PRINT "Debug on" ELSE PRINT "Debug off"
      CASE "X","x"
        kickConnection
      CASE "V","v"
        verbose = 1 - verbose
    END SELECT
    '
    IF LOC(#1) > 0 THEN
      IF showDebug THEN PRINT "*";
      x = waitFor(1000,150,"0,CONNECT","0,CLOSED","IPD,0,","IPD,1,","OK")
      IF showDebug THEN PRINT x," ",LEN(inTxt$)," ",inTxt$
      SELECT CASE x
        CASE 255 ' search strings not found before Max len reached
          IF inTxt$ = crlf$ THEN CONTINUE DO ' skip stray crlf$ pair
          IF showDebug THEN
            PRINT "<waiting for FTP action>"
            PRINT inTxt$
            PRINT"<END>"
          ENDIF
        CASE 0 ' timeout
          IF inTxt$ = crlf$ THEN CONTINUE DO ' skip stray crlf$ pair
          IF showDebug THEN
            PRINT "<timeout waiting for FTP action>"
            PRINT inTxt$
            PRINT"<END>"
          ENDIF
        CASE 1 ' 0,CONNECT
          doWelcome
          
        CASE 2 '0,CLOSED
          PRINT "Client disconnected"
        CASE 3 ' IPD,0, - FTP request coming
          CMDtxt$ = FTPcommand$()
          getCMD CMDtxt$
        CASE 4 ' IPD,1, - FTP data coming. we shouldn't see this here
          PRINT "request on 1 >";inTxt$
        CASE 5 ' OK
          CONTINUE DO
      END SELECT
    ENDIF
  LOOP UNTIL quit = 1
  
  PRINT "Shutting down..."
  ESPcommand "AT+CIPSTATUS"
  IF waitFor(2000,150,"STATUS:3") = 1 THEN
    ESPcommand "AT+CIPCLOSE=0"
  ENDIF
  ESPcommand("AT+CIPRECVMODE="+STR$(CIPRECVMODE))
  IF NOT OK() THEN PRINT " Error setting interactive data mode"
  CLOSE #1
  PAUSE 100
  CHDIR OldFolder$
  saveprefs "FTPserver.ini"
  PAUSE 50
END
  '
  
SUB getCMD CMDtxt$
  LOCAL a$, para$, type$
  LOCAL INTEGER x
  LOCAL FLOAT t
  a$ = CMDtxt$
  IF RIGHT$(a$,1) = CHR$(10) THEN a$ = LEFT$(a$,LEN(a$)-1)
  IF RIGHT$(a$,1) = CHR$(13) THEN a$ = LEFT$(a$,LEN(a$)-1)
  IF showDebug THEN PRINT LEN(a$)," ",a$ 'DEBUG
  IF LEN(a$) > 1 THEN
    CMDx$ = FIELD$(a$,1," ")
    IF showDebug THEN PRINT LEN(a$)," ",a$
    IF verbose THEN PRINT CMDx$
    SELECT CASE CMDx$
      CASE "FEAT"
        x = FTPresponce(features$)
      CASE "SYST"
        x = FTPresponce(systype$)
      CASE "USER"
        user$ = FIELD$(a$,2," ")
        x = FTPresponce("331 User "+user$+" OK. Password required"+crlf$)
      CASE "PASS"
        x = FTPresponce(passOK$)
      CASE "PASV"
        x = FTPresponce(unknownCMD$) 'passive$
      CASE "PORT"
        para$ = FIELD$(a$,2," ")
        clientIP$ = FIELD$(para$,1,",")+"."+FIELD$(para$,2,",")+"."+FIELD$(para$,3,",")+"."+FIELD$(para$,4,",")
        clientPort = VAL(FIELD$(para$,5,","))*256+VAL(FIELD$(para$,6,","))
        IF showDebug THEN PRINT clientIP$," ",clientPort 'DEBUG
        ESPcommand("AT+CIPSTART=1,"+qt$+"TCP"+qt$+","+qt$+clientIP$+qt$+","+STR$(clientPort))
        IF NOT OK() THEN
          ESPcommand("AT+CIPCLOSE=1")
          x = OK()
          PAUSE 500
          ESPcommand("AT+CIPSTART=1,"+qt$+"TCP"+qt$+","+qt$+clientIP$+qt$+","+STR$(clientPort))
          IF NOT OK() THEN
            PRINT " Error opening port for data mode"
          ENDIF
        ENDIF
        ' print inTxt$; 'DEBUG
        PAUSE 20
        x = FTPresponce(activePortOK$)
      CASE "LIST"
        x = FTPresponce(acceptTF$)
        PAUSE 10
        filecount = filelist() ' assemble file list and send on port 1
        x = FTPresponce(finishList$+STR$(filecount)+" matches total"+crlf$)
        IF verbose THEN PRINT STR$(filecount)+" files listed"
      CASE "DELE" ' delete file
        para$ = FIELD$(a$,2," ")
        IF para$<>"" THEN
          KILL para$
          x = FTPresponce("250 "+qt$+para$+qt$+" deleted"+crlf$)
        ENDIF
      CASE "PWD"
        para$ = CWD$
        IF RIGHT$(para$,1)<>"/" AND RIGHT$(para$,1)<>"\" THEN para$ = para$+"/"
        x = FTPresponce("257 "+qt$+para$+qt$+" is current directory"+crlf$)
      CASE "CWD"
        para$ = FIELD$(a$,2," ")
        CHDIR para$
        para$ = CWD$
        IF RIGHT$(para$,1)<>"/" AND RIGHT$(para$,1)<>"\" THEN para$ = para$+"/"
        x = FTPresponce("257 "+qt$+para$+qt$+" is current directory"+crlf$)
      CASE "CDUP" ' up one level
        CHDIR ".."
        para$ = CWD$
        IF RIGHT$(para$,1)<>"/" AND RIGHT$(para$,1)<>"\" THEN para$ = para$+"/"
        x = FTPresponce("257 "+qt$+para$+qt$+" is current directory"+crlf$)
      CASE "RETR" ' retrieve file
        para$ = FIELD$(a$,2," ")
        IF para$<>"" THEN
          x = FTPresponce(startTF$)
          startTime = TIMER
          IF verbose THEN PRINT "sending: ";para$
          x = sendFile(para$)
          t = (TIMER-startTime)/1000
          IF verbose THEN PRINT "send concluded. ";x;" bytes in ";STR$(t,2,2);" secs,";INT(x/t);" bytes/second."
          PAUSE 30
          x = FTPresponce(successTF$)
        ENDIF
      CASE "SIZE"
        para$ = FIELD$(a$,2," ")
        x = MM.INFO(FILESIZE para$)
        'print para$," ",x
        x = FTPresponce("213 "+STR$(x)+crlf$)
      CASE "TYPE"
        type$ = FIELD$(a$,2," ")
        IF type$ = "A" THEN
          x = FTPresponce(nowASCII$)
        ELSEIF type$ = "I" THEN
          x = FTPresponce(nowBINARY$)
        ELSE
          x = FTPresponce(unknownCMD$)
        ENDIF
      CASE "STOR" ' receive file
        para$ = FIELD$(a$,2," ")
        IF para$<>"" THEN
          PAUSE 20
          x = FTPresponce(startTF$)
          startTime = TIMER
          IF verbose THEN PRINT "Receiving: ";para$
          x = recFile(para$)
          t = (TIMER-startTime)/1000
          IF verbose THEN PRINT "Receive concluded, ";x;" bytes in ";STR$(t,2,2);" secs,";INT(x/t);" bytes/second."
          'pause 50
          x = FTPresponce(successTF$)
        ENDIF
      CASE "SITE"
        para$ = FIELD$(a$,2," ")
        IF para$ = "CHMOD" THEN
          x = FTPresponce(successTF$)
        ELSE
          x = FTPresponce(unknownCMD$)
        ENDIF
        
      CASE "MKD"
        para$ = FIELD$(a$,2," ")
        MKDIR para$
        x = FTPresponce(mkdirOK$)
      CASE "RNFR" ' rename from
        para2$ = FIELD$(a$,2," ")
        x = FTPresponce(nameFrom$)
      CASE "RNTO" ' rename to
        para$ = FIELD$(a$,2," ")
        IF verbose THEN PRINT para2$;" to  ";para$
        RENAME para2$ AS para$
        x = FTPresponce(nameto$)
      CASE "QUIT"
        quitting = 1
        x = FTPresponce(goodbye$)
        quitting = 0
      CASE ELSE
        x = FTPresponce(unknownCMD$)
    END SELECT
  ENDIF
END SUB
  
  'send ESP related command including 'request to send data'
SUB ESPcommand(txt$)
  flushInput
  PRINT #1,txt$+crlf$;
  IF showDebug THEN PRINT txt$ 'DEBUG
END SUB
  
SUB flushInput
  LOCAL junk$
  IF LOC(#1) > 0 THEN
    DO
      junk$ = INPUT$(255,#1)
    LOOP UNTIL LOC(#1) = 0
  ENDIF
END SUB
  
  ' wait timeout% mS or 250 bytes for "OK"
FUNCTION OK(timeout%)
  IF timeout% = 0 THEN timeout% = 500
  IF waitFor(timeout%,250,"OK") = 1 THEN OK = 1
END FUNCTION
  
  ' wait for text string(s). tested in supplied order and spips after first blank text
  ' returns index to found txt or 0 if timeout
  ' if maxlen is non zero, function will return 255 if max len is reached
  ' and global variable inTxt$ will contain the received text.
FUNCTION waitFor(timeout%,maxlen%,a1$,a2$,a3$,a4$,a5$,a6$) AS INTEGER
  LOCAL INTEGER n, c, endTime = TIMER + timeout%
  LOCAL a$
  DO
    IF LOC(#1) THEN
      a$ = a$ + INPUT$(1,#1)
      c = c + 1
      IF c >= maxlen% THEN n = 255 : EXIT DO
      IF RIGHT$(a$,LEN(a1$)) = a1$ THEN n = 1 : EXIT DO
      IF a2$="" THEN CONTINUE DO
      IF RIGHT$(a$,LEN(a2$)) = a2$ THEN n = 2 : EXIT DO
      IF a3$="" THEN CONTINUE DO
      IF RIGHT$(a$,LEN(a3$)) = a3$ THEN n = 3 : EXIT DO
      IF a4$="" THEN CONTINUE DO
      IF RIGHT$(a$,LEN(a4$)) = a4$ THEN n = 4 : EXIT DO
      IF a5$="" THEN CONTINUE DO
      IF RIGHT$(a$,LEN(a5$)) = a5$ THEN n = 5 : EXIT DO
      IF a6$="" THEN CONTINUE DO
      IF RIGHT$(a$,LEN(a6$)) = a6$ THEN n = 6 : EXIT DO
    ENDIF
  LOOP UNTIL TIMER > endtime OR c > 254
  IF maxlen% THEN inTxt$ = a$
  waitFor = n
END FUNCTION
  
  ' send responce to FTP command (IPD 0)
FUNCTION FTPresponce(txt$) AS INTEGER
  LOCAL b$
  LOCAL INTEGER x
  b$="AT+CIPSEND="+"0,"+STR$(LEN(txt$))
  ESPCommand b$
  x = waitFor(1000,60,">","ERROR")
  IF x = 1 THEN
    PRINT #1, txt$;
    IF waitFor(2000,60,"bytes") <> 1 AND quitting = 0 THEN PRINT "2 Error sending responce"
  ELSEIF x = 2 THEN
    PRINT "1 Error sending responce"
  ENDIF
  FTPresponce = x
END FUNCTION
  
  'convert received text into FTP command
FUNCTION FTPcommand$()
  LOCAL INTEGER x, k, i
  LOCAL txt$
  x = waitFor(500,100,crlf$)
  x = VAL(inTxt$) ' the number of characters in the buffer
  ESPCommand "AT+CIPRECVDATA=0,"+STR$(x)
  PAUSE 20
  x = waitFor(1000,100,"DATA:","DATA,") ' decode either ESP8266 or ESP32
  IF x = 1 OR x = 2 THEN
    x = waitFor(1000,50,",",":")
    IF x = 1 OR x = 2 THEN i = VAL(inTxt$)
  ENDIF
  IF x = 1 OR x = 2 THEN
    txt$ = INPUT$(i,#1)
    IF RIGHT$(txt$,1) = CHR$(10) THEN txt$ = LEFT$(txt$,LEN(txt$)-1)
    IF RIGHT$(txt$,1) = CHR$(13) THEN txt$ = LEFT$(txt$,LEN(txt$)-1)
    'print "*";txt$;"*"
    FTPcommand$ = txt$
  ENDIF
END FUNCTION
  
  ' send FTP data (IPD 1)in obuff()
SUB FTPsendData a$
  LOCAL b$, c$
  LOCAL INTEGER k, x, z
  IF LEN(a$)+LLEN(obuff())<= MTU THEN 'keep the MTU down
    LONGSTRING APPEND obuff(),a$
  ELSE 'too big to fit so send the current output buffer
    
    DO
      z = 0
      PAUSE 10
      b$="AT+CIPSEND=1,"+STR$(LLEN(obuff()))
      ESPCommand b$
      x = waitFor(1000,50,">","ERROR"+crlf$)
      IF x = 1 THEN
        PAUSE 10
        'longstring print #1, obuff();
        dripfeed
        IF verbose THEN
          pkt = pkt + 1
          IF (pkt MOD 10) = 0 THEN  PRINT ".";
          IF (pkt MOD 800) = 0 THEN  PRINT ""
        ENDIF
        k = waitFor(12000,50,"SEND OK","busy")
        IF k = 2 THEN
          PRINT "busy sending data"
          PAUSE 5000
          k = waitFor(5000,50,"SEND OK","busy")
          IF k = 1 THEN PRINT "Going again!"
        ELSEIF k <>1 THEN
          '
        ENDIF
      ELSEIF x = 2 THEN
        PRINT "1 Error sending data"
        pkt = pkt - 1
        z = 1
      ELSE
        PRINT "x";
        z = 1
      ENDIF
    LOOP UNTIL z = 0
    LONGSTRING CLEAR obuff()
    LONGSTRING APPEND obuff(),a$
    
  ENDIF
END SUB
  
  ' send the final data packet and close the chanel to indicate data finished.
SUB FTPendData
  LOCAL b$
  LOCAL INTEGER x ,k
  PAUSE 10
  b$="AT+CIPSEND=1,"+STR$(LLEN(obuff()))
  ESPCommand b$
  x = waitFor(500,30,">","ERROR")
  IF x = 1 THEN
    PAUSE 10
    'longstring print #1, obuff();
    dripfeed
    k = waitFor(5000,50,"SEND OK","busy")
    IF k = 2 THEN
      PRINT "busy sending data"
      PAUSE 5000
      k = waitFor(5000,50,"SEND OK","busy")
      IF k = 1 THEN PRINT "Going again!"
    ELSEIF k <>1 THEN
      'print k; ';k
    ENDIF
  ELSEIF x = 2 THEN
    PRINT "1 Error sending data"
  ENDIF
  LONGSTRING CLEAR obuff()
  IF verbose THEN PRINT ""
  pkt = 0
  PAUSE 20
  ESPCommand "AT+CIPCLOSE=1"
  IF NOT OK() THEN PRINT "Error finalising data send"
  
END SUB
  
  ' send data slowly, checking handshaking line for each byte.
SUB dripfeed
  LOCAL INTEGER n,k
  LOCAL ch$
  k = LLEN(obuff())
  FOR n = 1 TO k
    ch$ = LGETSTR$(obuff(), n, 1)
    DO
      IF NOT PIN(RTSpin) AND LOF(#1) > 255 THEN EXIT DO
    LOOP
    PRINT #1,ch$;
  NEXT n
END SUB
  
  ' sign on text send when FTP client first connects.
SUB dowelcome
  LOCAL INTEGER x
  x = FTPresponce(welcomeTxt$)
END SUB
  
  ' shows some info and a brief help text
SUB showStatus
  PRINT
  showTime
  ESPcommand "AT+CIPSTATUS"
  IF OK(3000) THEN
    IF INSTR(inTxt$,"STATUS:3") THEN
      PRINT "Local IP "+stationIP$
      PRINT "remote IP "+clientIP$+" Port: "+STR$(clientPort)
    ELSEIF INSTR(inTxt$,"STATUS:0") THEN
      PRINT "Not connected"
    ELSEIF INSTR(inTxt$,"STATUS:1") THEN
      PRINT "Inactive"
    ELSEIF INSTR(inTxt$,"STATUS:2") THEN
      PRINT "Connected to WiFi and waiting on IP "+stationIP$
    ENDIF
  ENDIF
  PRINT
  PRINT "<ESC> = quit , X = kick connection, D = toggle Debug, V = toggle Verbose, H = show status"
  PRINT "B = change baudrate, R = reset ESP"
END SUB
  
  ' use this to force close any hung FTP clcient connection
SUB kickConnection
  ESPcommand "AT+CIPCLOSE=0"
  IF OK() THEN
    PRINT " Client released"
  ELSE
    PRINT " Error releasing client"
  ENDIF
END SUB
  
  ' list files in current directory and send to client
FUNCTION filelist() AS INTEGER
  LOCAL f$, filespec$
  LOCAL INTEGER mf
  mf = 0
  
  f$ = DIR$("*", DIR)
  DO WHILE f$ <> ""
    filespec$ = LISTdate$(f$,0)
    FTPsendData filespec$+f$+crlf$
    mf = mf + 1
    f$ = DIR$()
  LOOP
  f$ = DIR$("*", FILE)
  DO WHILE f$ <> ""
    filespec$ = LISTdate$(f$,1)
    FTPsendData filespec$+f$+crlf$
    mf = mf + 1
    f$ = DIR$()
  LOOP
  FTPendData
  filelist = mf
END FUNCTION
  
  ' format file date to siuit FTP expectations
FUNCTION LISTdate$(f$,type AS INTEGER)
  LOCAL fdate$, fdata$, fspec$, sm$ = "  JanFebMarAprMayJunJulAugSepOctNovDec"
  LOCAL INTEGER m, ep,epf
  ep = EPOCH(now)
  
  fdate$=MM.INFO(MODIFIED f$)
  fdata$ = MID$(fdate$,9,2)+MID$(fdate$,5,4)+MID$(fdate$,1,4)+MID$(fdate$,11,8)
  epf = EPOCH(fdata$)
  m = VAL(MID$(fdate$,6,2))
  IF ep - epf > 28512000 THEN ' more than 11 months old so use year instead of time
    fspec$ = MID$(sm$,m*3,3)+" "+MID$(fdate$,9,2)+"  "+MID$(fdate$,1,4)+" "
  ELSE
    fspec$ = MID$(sm$,m*3,3)+" "+MID$(fdate$,9,2)+" "+MID$(fdate$,12,5)+" "
  ENDIF
  
  IF type = 0 THEN ' directory
    LISTdate$ = "drwxrwxr-x    2 "+user$+" "+user$+"              0 "+fspec$
  ELSE
    m = MM.INFO(FILESIZE f$)
    fdata$ = STR$(m,15,0," ")+" "
    LISTdate$ = "-rw-rw-r--    1 "+user$+" "+user$+" "+fdata$+fspec$
  ENDIF
END FUNCTION
  
  ' receive a file
FUNCTION recFile(f$) AS INTEGER
  LOCAL a$
  LOCAL INTEGER x, k, j,i, s, lastpacket
  LOCAL INTEGER endTime = TIMER + 600000 ' allow 10 minutes for large files.
  OPEN f$ FOR OUTPUT AS #4
  PAUSE 50
  DO
    x = waitFor(1000,100,"+IPD,1,","OK",",CLOSED","+IPD,0,","ERROR","busy ")
    IF showDebug THEN PRINT inTxt$
    SELECT CASE x
      CASE 0,4,255
        'print x
        EXIT DO
      CASE 3
        PRINT "c"
        EXIT DO
      CASE 2, 5
        'print x
        CONTINUE DO
      CASE 6
        PAUSE 3000
        CONTINUE DO
      CASE 1
        i = waitFor(500,100,crlf$)
        i = VAL(inTxt$) ' the number of characters in the buffer
        inTxt$ = inTxt$ + INPUT$(50,#1)
        IF INSTR(inTxt$,"1,CLOSED") > 0 THEN lastpacket = 1 'ESP8266 sends this too soon
        ESPCommand "AT+CIPRECVDATA=1,"+STR$(i)
        x = waitFor(1000,100,"DATA:","DATA,","ERROR")
        IF x = 3 THEN ' try again
          PAUSE 20
          ESPCommand "AT+CIPRECVDATA=1,"+STR$(i)
          x = waitFor(1000,100,"DATA:","DATA,","ERROR")
          IF x = 3 THEN CONTINUE DO
        ENDIF
        
        IF x = 1 OR x = 2 THEN
          x = waitFor(1000,50,",",":")
          IF x = 1 OR x = 2 THEN i = VAL(inTxt$)
        ENDIF
        IF x = 1 OR x = 2 THEN
          j = MIN(i,255)
          a$=INPUT$(j,#1)
          s = s + i
          i = i-LEN(a$)
          '          if left$(a$,4)=crlf$+"OK" then a$ = mid$(a$,5) ; print "OK at start" ' checking for stray <CRLF>OK
          PRINT #4, a$;
          
          IF i > 0 THEN
            IF verbose THEN
              k = k + 1
              IF (k MOD 10) = 0 THEN  PRINT ".";
              IF (k MOD 800) = 0 THEN  PRINT ""
            ENDIF
            DO
              IF showDebug THEN PRINT i
              j = MIN(i,255)
              a$=INPUT$(j,#1)
              IF a$ <>"" THEN
                i = i-LEN(a$)
                '                if i = 0 then
                '                  if right$(a$,4) = crlf$+"OK" then
                '                    a$ = left$(a$,len(a$)-4)
                '                    print "OK at end ";s ' checking for stray <CRLF>OK
                '                  endif
                '                endif
                PRINT #4, a$;
                
              ENDIF
              'pause 10
            LOOP UNTIL i = 0
          ENDIF
        ENDIF
    END SELECT
    PAUSE 5
    IF lastpacket = 1 THEN EXIT DO
  LOOP UNTIL TIMER >= endTime
  PRINT ""
  PAUSE 10
  CLOSE #4
  PAUSE 200
  recFile = s
END FUNCTION
  
  ' send a file
FUNCTION sendfile(f$) AS INTEGER
  LOCAL INTEGER m, x
  LOCAL a$
  m = MM.INFO(FILESIZE f$)
  LONGSTRING CLEAR obuff()
  IF m > 0 THEN
    OPEN f$ FOR INPUT AS #4
    SEEK #4,1
    DO
      a$ = INPUT$(255,#4)
      IF a$<>"" THEN FTPsendData(a$)
    LOOP UNTIL EOF(#4)
    CLOSE #4
    PRINT ""
    FTPendData
  ENDIF
  SendFile = m
END FUNCTION
  
  ' set CMM2 time and date from NTP provided year is not 1970 (time not set)
SUB settimefromNTP
  LOCAL NTPhour$, NTPdate$, sm$ = "  JanFebMarAprMayJunJulAugSepOctNovDec"
  LOCAL INTEGER x
  ESPcommand "AT+CIPSNTPTIME?"
  IF OK() THEN
    x = INSTR(inTxt$,"TIME:")
    IF x THEN
      ntpTime$ = MID$(inTxt$,x+5,24)
      NTPhour$ = MID$(ntpTime$,12,8)
      NTPdate$ = MID$(ntpTime$,9,2)+"-"+STR$(INSTR(sm$,MID$(ntpTime$,5,3))/3,2,0,"0")+"-"+MID$(ntpTime$,21,4)
      PRINT NTPDate$;" ";NTPhour$
      IF RIGHT$(NTPdate$,4) = "1970" THEN
        PRINT "Time not set!"
      ELSE
        PRINT "CMM2 time updated."
        ' here is where we will do the update
        TIME$ = NTPhour$
        DATE$ = NTPdate$
      ENDIF
    ENDIF
  ENDIF
END SUB
  
  ' prin the time in NTP format
SUB showTime
  LOCAL INTEGER x
  ESPcommand "AT+CIPSNTPTIME?"
  IF OK() THEN
    x = INSTR(inTxt$,"TIME:")
    IF x THEN
      ntpTime$ = MID$(inTxt$,x+5,24)
      PRINT ntpTime$
    ENDIF
  ENDIF
END SUB
  
  ' reset the ESP
FUNCTION ESPreset()
  ESPcommand("AT+RST")
  ESPreset = waitFor(20000,150,"WIFI GOT IP")
END FUNCTION
  
  ' called on startup or after reset
SUB ESPstart
  ESPcommand("AT+CIFSR")
  IF NOT OK() THEN PRINT " Error getting IP"
  x = INSTR(inTxt$,"+CIFSR:STAIP")
  IF x > 0 THEN
    stationIP$ = MID$(inTxt$,x,35)
    stationIP$ = FIELD$(stationIP$,2,qt$)
  ENDIF
  ESPcommand "AT+CIPSNTPCFG=1,"+NTPdata$
  IF NOT OK() THEN PRINT " Error setting NTP servers"
  
  ESPcommand("AT+CIPRECVMODE?")
  IF OK() THEN
    x = INSTR(inTxt$,"MODE:")
    IF x THEN CIPRECVMODE = VAL(MID$(inTxt$,x+6,1)) ' save setting so we can restore on program exit
  ENDIF
  
  ESPcommand("AT+CIPRECVMODE=1")
  IF NOT OK() THEN PRINT " Error setting interactive data mode"
  
  ESPcommand("AT+CIPMUX=1")
  IF NOT OK() THEN PRINT " Error setting MUX"
  
  ESPcommand("AT+CIPSERVER=0")
  IF NOT OK() THEN PRINT " Error killing Server"
  
  ESPcommand("AT+CIPSERVERMAXCONN=1")
  IF NOT OK() THEN PRINT " Error setting Server limit"
  
  ESPcommand("AT+CIPSERVER=1,21")
  IF NOT OK() THEN PRINT " Error setting Server"
  
  ESPcommand("AT+CIPSTO="+FTPtimeout$)
  IF NOT OK() THEN PRINT " Error setting timeout"
  PRINT "Nearly ready"
END SUB
  
  ' immediate change of baudrate. New setting gets saved to ini file on shutdown
SUB changeBaud
  LOCAL validsp$ = "1=38400 2=57600 3=115200 4=230400 5=460800 6=921600 "
  LOCAL INTEGER x
  PRINT
  PRINT "Current baud rate = ";ComSpeed$
  PRINT validsp$
  INPUT "Choose the new speed: "; x
  IF x > 0 AND x < 7 THEN
    ComSpeed$ = MID$(FIELD$(validSp$,x," "),3)
    prefs$(findPref("comSpeed"),2) = ComSpeed$
    ESPcommand "AT+UART_DEF="+ComSpeed$+",8,1,0,1"
    IF OK() THEN
      CLOSE #1
      PAUSE 100
      OPEN comPort$+ComSpeed$+",8192" AS #1
      PRINT "Com port set to: ";comPort$+ComSpeed$+",8192"
    ENDIF
    ESPcommand "AT+UART_DEF?"
    IF OK() THEN 
    IF showDebug THEN PRINT inTxt$
    ENDIF
  ENDIF
END SUB
  
' read the pref$() array and setup the varaibles
SUB configure
  LOCAL txt1$,txt2$,txt3$
  ComPort$ = prefs$(findPref("comPort"),2)
  IF RIGHT$(ComPort$,1) <>":" THEN ComPort$ = ComPort$+":"
  ComSpeed$ = prefs$(findPref("comSpeed"),2)
  ESP32 = VAL(prefs$(findPref("ESP32"),2))
  doReset = VAL(prefs$(findPref("Reset"),2))
  MTU = VAL(prefs$(findPref("MTU"),2))
  MTU = MIN(MTU,2048)
  RTSpin = VAL(prefs$(findPref("RTSpin"),2))
  SSID$=qt$+prefs$(findPref("AccessPoint"),2)+qt$
  SSIDPassword$=qt$+prefs$(findPref("APpassword"),2)+qt$
  txt1$ = prefs$(findPref("NTPserver1"),2)
  txt2$ = prefs$(findPref("NTPserver2"),2)
  txt3$ = prefs$(findPref("NTPserver3"),2)
  NTPdata$ = prefs$(findPref("timezone"),2)+","+qt$+txt1$+qt$+","+qt$+txt2$+qt$+","+qt$+txt3$+qt$
  FTPtimeout$ = prefs$(findPref("FTPtimeout"),2)
END SUB
  
  ' default configuration data
  ' this will get used if the data is not in FTPserver.ini or the file does not exist.
  ' change these to suit
  ' access point name and pw not currently used
prefData:
  DATA "comPort = com1:"
  DATA "comSpeed = 115200"
  DATA "ESP32 = 1"
  DATA "Reset = 0"
  DATA "MTU = 2048" ' 2048 is the maximum allowed
  DATA "RTSpin = 36"
  DATA "AccessPoint = MYWIFI"
  DATA "APpassword = MYPWD"
  DATA "timezone = 11"
  DATA "NTPserver1 = 10.1.1.52"
  DATA "NTPserver2 = au.pool.ntp.org"
  DATA "NTPserver3 = 10.1.1.1"
  DATA "FTPtimeout = 1800"
  DATA "EOF"
  
  ' read the ini file and fill the prefs$() array
SUB openprefs prefsFile$
  LOCAL txt$, pref$, prefdata$, f$
  LOCAL INTEGER j, n = BOUND(prefs$(),1)
  RESTORE prefData
  FOR j = 1 TO n' load the default values
    READ txt$
    IF txt$ = "EOF" THEN EXIT FOR
    pref$ = trim$(FIELD$(txt$,1,"="))
    prefdata$ = trim$(FIELD$(txt$,2,"="))
    prefs$(j,1) = pref$
    prefs$(j,2) = prefdata$
  NEXT j
  
  f$ = DIR$(prefsFile$, FILE) ' if ini file exists, read and update values
  IF f$ <> "" THEN
    OPEN prefsFile$ FOR INPUT AS #3
    DO
      LINE INPUT #3,txt$
      pref$ = trim$(FIELD$(txt$,1,"="))
      prefdata$ = trim$(FIELD$(txt$,2,"="))
      FOR j = 1 TO n
        IF prefs$(j,1) = pref$ THEN prefs$(j,2) = prefdata$
      NEXT j
    LOOP UNTIL EOF(#3)
    CLOSE #3
  ENDIF
END SUB
  
  ' save all data in prefs$() to ini file
SUB saveprefs prefsFile$
  LOCAL txt$
  LOCAL INTEGER j, n = BOUND(prefs$(),1)
  OPEN prefsFile$ FOR OUTPUT AS #3
  FOR j = 1 TO n
    IF prefs$(j,1) = "" THEN EXIT FOR
    txt$ = prefs$(j,1)+" = "+prefs$(j,2)
    PRINT #3,txt$
  NEXT j
  CLOSE #3
END SUB
  
  ' search the prefs$() array for a matching parameter name.
FUNCTION findPref(txt$)
  LOCAL INTEGER j, n = BOUND(prefs$(),1)
  FOR j = 1 TO n
    IF UCASE$(txt$) = UCASE$(prefs$(j,1)) THEN
      findPref = j
      EXIT FOR
    ENDIF
  NEXT j
END FUNCTION
  
FUNCTION trim$(txt$)
  trim$ = ltrim$(rtrim$(txt$))
END FUNCTION
  
FUNCTION LTrim$(a$)
  LOCAL n
  IF LEN(a$)=0 THEN
    LTrim$ = ""
  ELSE
    FOR n= 1 TO LEN(a$)
      IF MID$(a$,n,1)<>" " AND MID$(a$,n,1)<>CHR$(9) THEN EXIT FOR
    NEXT n
    LTrim$=MID$(a$,n)
  ENDIF
END FUNCTION
  
FUNCTION RTrim$(a$)
  LOCAL n
  IF LEN(a$)=0 THEN
    RTrim$ = ""
  ELSE
    FOR n= LEN(a$) TO 1 STEP -1
      IF MID$(a$,n,1)<>" " AND MID$(a$,n,1)<>CHR$(9) THEN EXIT FOR
    NEXT n
    RTrim$=LEFT$(a$,n)
  ENDIF
END FUNCTION